// Client, client application for the serialmon board.
//
// Copyright (C) 2008, Frank Tkalcevic, www.franksworkshop.com

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace Client
{
    abstract class DisplayDataCollection
    {
        public abstract DisplayData this[int n] { get; set; }
        public abstract int Length { get;}
    }

    class DisplayDataArray : DisplayDataCollection
    {
        DisplayData[] data;

        public DisplayDataArray(int nSize)
        {
            data = new DisplayData[nSize];
        }

        public override DisplayData this[int n] 
        {
            get { return data[n]; }
            set { data[n] = value;  }
        }

        public override int Length 
        {
            get { return data.Length; }
        }
    }

    class DisplayDataList : DisplayDataCollection
    {
        List<DisplayData> data;

        public DisplayDataList()
        {
            data = new List<DisplayData>();
        }

        public override DisplayData this[int n]
        {
            get { return data[n]; }
            set 
            {
                if (n == data.Count)
                    data.Add(value);
                else
                    data[n] = value; 
            }
        }

        public override int Length
        {
            get { return data.Count; }
        }
    }

    abstract class DisplayMode
    {
        protected int nDistanceFromStart;
        protected int nNextCharPosition;
        protected int nNextLinePosition;
        protected int nBufferLines;
        protected int nWidestLine;
        protected CircularBuffer<ushort> DataBuffer;
        protected List<DisplayDataCollection> DisplayBuffer;
        protected DisplayProperties DisplayProps;

        protected abstract void AddText(string s);
        protected abstract void AddData( int nLine, char c);
        protected abstract void AddTime(byte nCount);

        public DisplayMode(CircularBuffer<ushort> buf, DisplayProperties props)
        {
            nBufferLines = 0;
            nWidestLine = 0;
            nDistanceFromStart = 0;
            nNextLinePosition = 0;
            nNextCharPosition = 0;
            DataBuffer = buf;
            DisplayProps = props;
            DisplayBuffer = new List<DisplayDataCollection>();
        }

        public virtual void UpdateScrollRanges(ushort data)
        {
            SerialCommand.ECommand cmd = SerialCommand.Command(data);

            if (cmd == SerialCommand.ECommand.Data)
            {
                if (DisplayProps.RealTime)
                {
                    byte nTime = (byte)(data >> 12);
                    if (nTime > 0)
                        AddTime(nTime);
                }
                AddData(SerialCommand.Line(data), (char)(data & 0xFF));
            }
            else if (cmd == SerialCommand.ECommand.Timing && DisplayProps.RealTime)
            {
                AddTime((byte)(data & 0xFF));
            }

            nWidestLine = DisplayProps.CharactersPerLine;
        }

        public virtual void RecomputeScrollRanges()
        {
            nDistanceFromStart = 0;
            nNextCharPosition = 0;
            nNextLinePosition = 0;
            nWidestLine = 0;
            DisplayBuffer = new List<DisplayDataCollection>();

            for (int i = 0; i < DataBuffer.Count; i++)
            {
                ushort data = DataBuffer[i];
                UpdateScrollRanges(data);
            }
        }

        public virtual int BufferLines
        {
            get { return DisplayBuffer.Count; }
        }

        public virtual int WidestLine
        {
            get { return nWidestLine; }
        }

        static public bool IsPrint(int n)
        {
            if (n >= 0x20 && n <= 0x7f)
                return true;
            else
                return false;
        }

        public List<DisplayDataCollection> Display
        {
            get { return this.DisplayBuffer; }
        }

    }
}
