// Client, client application for the serialmon board.
//
// Copyright (C) 2008, Frank Tkalcevic, www.franksworkshop.com

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// Example - http://www.beyondlogic.org/protocolanalyser/protocolanalyser.htm

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Microsoft.Win32;


namespace Client
{
    public partial class MainForm : Form
    {
        const int MarginY = 5;
        const int MarginX = 5;
        private const string ApplicationKey = @"HKEY_CURRENT_USER\Software\FranksWorkshop\SerialMonitor";

        SerialPortEx Port;
        // todo - make this configurable - or better yet, no max
        CircularBuffer<ushort> DataBuffer = new CircularBuffer<ushort>( 500000);        // Hard coded buffer max at 1MB
        Rectangle rcDisplay;

        static Color clrBackground = SystemColors.Window;
        Brush brBackground = new SolidBrush(clrBackground);

        int TextFontWidth = 10;
        int TextFontHeight = 20;

        bool bRunning = false;
        int nFirstVisible = 0;
        int nAutoBaudingLine = 0;
        bool bAutoBauding = true;
        String sSerialMonPort;

        List<byte> buf = null;

        DisplayMode displayMode = null;
        DisplayProperties displayProps = new DisplayProperties();

        CLineProperties[] LineProperties;

        public MainForm()
        {
            LoadSettings();
            displayMode = new DisplayModeHex(DataBuffer, displayProps);

            LineProperties = new CLineProperties[2];
            LineProperties[0] = new CLineProperties(new Font(FontFamily.GenericMonospace, 12, FontStyle.Regular), SystemColors.WindowText, Color.PaleGoldenrod);
            LineProperties[1] = new CLineProperties(new Font(FontFamily.GenericMonospace, 12, FontStyle.Regular), SystemColors.WindowText, Color.PaleGreen);

            InitializeComponent();

            displayProps.ShowLine1 = chkLine1.Checked;
            displayProps.ShowLine2 = chkLine2.Checked;

            groupBoxLine1.BackColor = Color.PaleGoldenrod;
            groupBoxLine2.BackColor = Color.PaleGreen;
            radioButtonHex.Checked = true;

            Port = new SerialPortEx();
            Port.PortName = sSerialMonPort;
            Port.BaudRate = 250000;
            Port.DataBits = 8;
            Port.Handshake = System.IO.Ports.Handshake.None;
            Port.Parity = System.IO.Ports.Parity.None;
            Port.StopBits = System.IO.Ports.StopBits.One;
            Port.ReadTimeout = 50;
            Port.DataReceived += new System.IO.Ports.SerialDataReceivedEventHandler(Port_DataReceived);
            Port.ErrorReceived += new System.IO.Ports.SerialErrorReceivedEventHandler(Port_ErrorReceived);
            Port.ReceivedBytesThreshold = 2;

            InitialiseSettings();
            RecomputeScrollRanges();
        }

        private SizeF GetTextSpacing(Font f)
        {
            int lineSpacing;        // font family line spacing in design units
            float lineSpacingPixel; // line spacing converted to pixels

            // Display the line spacing in design units and pixels.
            lineSpacing = f.FontFamily.GetLineSpacing(f.Style);
            lineSpacingPixel = f.Size * lineSpacing / f.FontFamily.GetEmHeight(f.Style);
            lineSpacingPixel *= 1.5f;

            string s = "0123";
            SizeF sz = this.CreateGraphics().MeasureString(s, f );
            sz.Width /= s.Length;
            sz.Height = lineSpacingPixel;

            return sz;
        }

        void InitialiseSettings()
        {
            cboBaud1.Text = "Auto";
            cboDataBits1.Text = "8";
            cboParity1.Text = "None";
            cboStopBits1.Text = "1";

            cboBaud2.Text = "Auto";
            cboDataBits2.Text = "8";
            cboParity2.Text = "None";
            cboStopBits2.Text = "1";

        }

        void Port_ErrorReceived(object sender, System.IO.Ports.SerialErrorReceivedEventArgs e)
        {
            //throw new Exception("The method or operation is not implemented.");
            System.Diagnostics.Debug.WriteLine("Error=" + e.EventType.ToString());
        }


        enum EReadState
        {
            WaitForCmd,
            ReadData,
            ReadBaud
        };

        EReadState ReadState = EReadState.WaitForCmd;
        byte nData1=0;

        private delegate void UpdateDataDelegate(int nFirst, int nLast);
        private delegate void AutoBaudDelegate(int nLine, int nBaud);

        void Port_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            //int nBytes = Port.BytesToRead;
            //byte[] data = new byte[nBytes];
            //Port.Read(data, 0, nBytes);

            // Process the data.
            int nFirst = DataBuffer.Count;
            while (Port.BytesToRead > 0)
            {
                byte b = (byte)Port.ReadByte();

                switch (ReadState)
                {
                    case EReadState.WaitForCmd:
                        SerialCommand.ECommand nCmd = SerialCommand.Command( b );
                        if (nCmd == SerialCommand.ECommand.Data || nCmd == SerialCommand.ECommand.Timing)
                        {
                            nData1 = b;
                            ReadState = EReadState.ReadData;
                        }
                        else if (nCmd == SerialCommand.ECommand.AutoBaud)
                        {
                            nData1 = b;
                            ReadState = EReadState.ReadBaud;
                            buf = new List<byte>();
                        }
                        break;

                    case EReadState.ReadData:
                        ushort nData = (ushort)((nData1 << 8) | b);
                        DataBuffer.Add(nData);
                        ReadState = EReadState.WaitForCmd;
                        break;

                    case EReadState.ReadBaud:
                        buf.Add(b);
                        if (buf.Count == 4)
                        {
                            ReadState = EReadState.WaitForCmd;
                            int nBaud = BitConverter.ToInt32(buf.ToArray(), 0);
                            this.BeginInvoke(new AutoBaudDelegate(AutoBaudUpdate), new object[] { (nData1 >> 1) & 1, nBaud });
                        }
                        break;
                }
            }
            int nLast = DataBuffer.Count;
            this.BeginInvoke(new UpdateDataDelegate(NewDataSamples), new object[] { nFirst, nLast });
        }

        private void UpdateScrollRanges(ushort nData)
        {
            displayMode.UpdateScrollRanges(nData);
        }

        // Change of display modes
        private void RecomputeScrollRanges()
        {
            nFirstVisible = 0;
            displayMode.RecomputeScrollRanges();
            UpdateScrollBars();
        }

        private void UpdateScrollBars()
        {
            vScrollBar1.Maximum = displayMode.BufferLines;
            if (vScrollBar1.Value > vScrollBar1.Maximum)
                vScrollBar1.Value = vScrollBar1.Maximum;

            hScrollBar1.Maximum = displayMode.WidestLine;
            if (hScrollBar1.Value > hScrollBar1.Maximum)
                hScrollBar1.Value = hScrollBar1.Maximum;
        }

        private void NewDataSamples( int nFirst, int nLast )
        {
            bool bAtEnd = false;

            if (vScrollBar1.Value == vScrollBar1.Maximum - vScrollBar1.LargeChange)
                bAtEnd = true;

            for (int i = nFirst; i < nLast; i++)
            {
                // Update scrollbar ranges
                UpdateScrollRanges(DataBuffer[i]);
            }

            UpdateScrollBars();

            if (bAtEnd)
            {
                if (vScrollBar1.Maximum > displayProps.LinesPerPage)
                {
                    vScrollBar1.Value = vScrollBar1.Maximum - vScrollBar1.LargeChange;
                    ScrollTo(vScrollBar1.Value);
                }
            }

            //todo only invalidate dirty area.
            Invalidate(rcDisplay);
        }

        UInt32 GetBaud(int nLine)
        {
            ComboBox cbo = nLine == 0 ? cboBaud1 : cboBaud2;
            if (cbo.Text == "Auto")
                return 0;
            else
            {
                UInt32 ret = 0;
                UInt32.TryParse(cbo.Text, out ret);
                return ret;
            }
        }

        void SetBaud(int nLine, UInt32 nBaud)
        {
            ComboBox cbo = nLine == 0 ? cboBaud1 : cboBaud2;
            if (nBaud == 0)
                cbo.Text = "Auto";
            else
                cbo.Text = nBaud.ToString();
        }

        private void SetDisplayBaud()
        {
            UInt32 nBaud1 = GetBaud(0);
            UInt32 nBaud2 = GetBaud(1);

            displayProps.Baud = (double)Math.Max(nBaud1, nBaud2);
        }

        private void AutoBaudUpdate(int nLine, int nBaud)
        {
            //System.Diagnostics.Debug.Assert(nLine == nAutoBaudingLine);
            if (nLine != nAutoBaudingLine)
                return;

            if (nLine == 0)
            {
                cboBaud1.Text = nBaud.ToString();

                if (cboBaud2.Text == "Auto")
                {
                    AutoBaud(1);
                }
                else
                {
                    SerialMon.ConfigLine(Port, 1, int.Parse(cboBaud1.Text), 8, 0, 1);
                    SetDisplayBaud();
                    SerialMon.Start(Port);
                }
            }
            else
            {
                cboBaud2.Text = nBaud.ToString();
                SetDisplayBaud(); 
                SerialMon.Start(Port);
            }
        }

        private void AutoBaud(int nLine)
        {
            nAutoBaudingLine = nLine;
            bAutoBauding = true;
            SerialMon.AutoConfigLine(Port, nLine);
        }

        private void btnStart_Click(object sender, EventArgs e)
        {
            try
            {
                Port.Open();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Unable to open COM port to serial monitor -\r\n" + ex.Message);
                return;
            }

            bRunning = true;
            if (cboBaud1.Text == "Auto")
            {
                AutoBaud(0);
            }
            else
            {
                if (cboBaud2.Text != "Auto")
                {
                    SerialMon.ConfigLine(Port, 0, int.Parse(cboBaud1.Text), 8, 0, 1);
                    SetDisplayBaud(); 
                    SerialMon.Start(Port);
                }
            }
        }

        private void btnStop_Click(object sender, EventArgs e)
        {
            if (bRunning)
            {
                SerialMon.Stop(Port);
                Port.Close();
                bRunning = false;
            }
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);

            // Reposition the scrollbars
            hScrollBar1.Left = 0;
            hScrollBar1.Width = ClientRectangle.Width - vScrollBar1.Width;
            hScrollBar1.Top = ClientRectangle.Height - hScrollBar1.Height;

            vScrollBar1.Top = groupBoxLine1.Bottom + MarginY;
            vScrollBar1.Height = ClientRectangle.Height - vScrollBar1.Top - hScrollBar1.Height;
            vScrollBar1.Left = ClientRectangle.Width - vScrollBar1.Width;

            // recalculate the display area.
            rcDisplay.X = MarginX;
            rcDisplay.Y = groupBoxLine1.Bottom + MarginY;
            rcDisplay.Width = ClientRectangle.Width - rcDisplay.X - vScrollBar1.Width;
            rcDisplay.Height = ClientRectangle.Height - rcDisplay.Y - hScrollBar1.Height;

            SizeF sz0 = GetTextSpacing(LineProperties[0].TextFont);
            SizeF sz1 = GetTextSpacing(LineProperties[1].TextFont);
            TextFontWidth = (int)(Math.Max(sz0.Width,sz1.Width) + 1);
            TextFontHeight = (int)(Math.Max(sz0.Height, sz1.Height) + 1);

            displayProps.CharactersPerLine = (int)(rcDisplay.Width / TextFontWidth);
            displayProps.LinesPerPage = (int)(rcDisplay.Height / TextFontHeight);

            vScrollBar1.SmallChange = 1;
            vScrollBar1.LargeChange = displayProps.LinesPerPage;
            hScrollBar1.SmallChange = 1;
            hScrollBar1.LargeChange = displayProps.CharactersPerLine;

            RecomputeScrollRanges();
            Invalidate(rcDisplay);
        }

        public override Size MinimumSize
        {
            get
            {
                return new Size(476, 214);
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            this.DoubleBuffered = true;
            e.Graphics.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;

            // Erase the background
            e.Graphics.FillRectangle(brBackground, rcDisplay);

            int nDisplayLine = 0;
            for (int nLine = nFirstVisible; nLine < nFirstVisible + displayProps.LinesPerPage; nLine++, nDisplayLine++)
            {
                if (nLine >= displayMode.Display.Count)
                    break;

                DisplayDataCollection Line = displayMode.Display[nLine];
                for (int nChar = 0; nChar < Line.Length; nChar++)
                {
                    float x = (nChar - nHOffset) * TextFontWidth;
                    float y = nDisplayLine * TextFontHeight;
                    PointF pt = new PointF(rcDisplay.Left + x, rcDisplay.Top + y);

                    if (Line[nChar].type == EDisplayDataType.Time)
                    {
                        pt.X += TextFontWidth / 2;
                        pt.Y += TextFontHeight / 2;
                        Point ptI = Point.Round(pt);
                        e.Graphics.DrawLine(Pens.Black, ptI.X, ptI.Y, ptI.X + 1, ptI.Y + 1);
                    }
                    else if (Line[nChar].type == EDisplayDataType.Text)
                    {
                        string s = new string(Line[nChar].data, 1);
                        e.Graphics.DrawString(s, LineProperties[0].TextFont, Brushes.Black, pt);
                    }
                    else if ( Line[nChar].type != EDisplayDataType.None)
                    {
                        string s = new string(Line[nChar].data, 1);
                        int nDataLine = Line[nChar].type == EDisplayDataType.Line1 ? 0 : 1;
                        e.Graphics.FillRectangle(LineProperties[nDataLine].BackgroundBrush, pt.X, pt.Y, TextFontWidth, TextFontHeight);
                        e.Graphics.DrawString(s, LineProperties[nDataLine].TextFont, LineProperties[nDataLine].TextBrush, pt);
                    }
                }

            }
        }

        private void ScrollTo(int n)
        {
            nFirstVisible = n;
            Invalidate(rcDisplay);
            
        }

        private void vScrollBar1_Scroll(object sender, ScrollEventArgs e)
        {
            ScrollTo(e.NewValue);
        }

        int nHOffset = 0;
        private void hScrollBar1_Scroll(object sender, ScrollEventArgs e)
        {
            if (e.NewValue != e.OldValue)
            {
                nHOffset = e.NewValue;
                Invalidate(rcDisplay);
            }
        }

        private void Clear()
        {
            DataBuffer.Clear();
            nHOffset = 0;
            hScrollBar1.Minimum = hScrollBar1.Maximum = hScrollBar1.Value = 0;
            vScrollBar1.Minimum = vScrollBar1.Maximum = vScrollBar1.Value = 0;
            RecomputeScrollRanges();
            Invalidate(rcDisplay);
        }

        private void btnClear_Click(object sender, EventArgs e)
        {
            Clear();
        }

        private void mnuFileOpen_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();

            dlg.Filter = "trace files (*.trc)|*.trc|All files (*.*)|*.*";
            dlg.FilterIndex = 1;
            dlg.RestoreDirectory = true;
            dlg.CheckFileExists = true;
            dlg.Multiselect = false;
            dlg.ValidateNames = true;

            if (dlg.ShowDialog(this) == DialogResult.OK)
            {
                FileStream s = File.Open(dlg.FileName, FileMode.Open, FileAccess.Read, FileShare.Read);

                byte[] bauds = new byte[sizeof(UInt32) * 2];
                s.Read( bauds, 0, 2*sizeof(UInt32));
                UInt32 baud1 = BitConverter.ToUInt32(bauds, 0);
                UInt32 baud2 = BitConverter.ToUInt32(bauds, sizeof(UInt32));
                SetBaud(0, baud1);
                SetBaud(1, baud2);

                long nLen = s.Length - 2 * sizeof(UInt32);
                byte[] b = new byte[nLen];
                s.Read(b, 0, (int)nLen);
                s.Close();
                s.Dispose();

                Clear();
                for (int i = 0; i < nLen; i += 2)
                    DataBuffer.Add(BitConverter.ToUInt16(b, i));

                RecomputeScrollRanges();
                Invalidate(rcDisplay);
            }
        }

        private void mnuFileSave_Click(object sender, EventArgs e)
        {
            SaveFileDialog dlg = new SaveFileDialog();

            dlg.Filter = "trace files (*.trc)|*.trc|All files (*.*)|*.*"  ;
            dlg.FilterIndex = 1;
            dlg.RestoreDirectory = true;
            dlg.CheckPathExists = true;
            dlg.OverwritePrompt = true;

            if (dlg.ShowDialog(this) == DialogResult.OK)
            {
                ushort[] arr = DataBuffer.ToArray();
                byte[] b = new byte[arr.Length * sizeof(ushort)];
                for (int i = 0; i < arr.Length; i++)
                    BitConverter.GetBytes(arr[i]).CopyTo(b, i * sizeof(ushort));

                FileStream s = File.Open(dlg.FileName, FileMode.Create, FileAccess.Write, FileShare.None);
                UInt32 baud1 = GetBaud(0);
                UInt32 baud2 = GetBaud(1);
                s.Write(BitConverter.GetBytes(baud1), 0, sizeof(UInt32));
                s.Write(BitConverter.GetBytes(baud1), 0, sizeof(UInt32));
                s.Write(b, 0, b.Length);
                s.Close();
                s.Dispose();
            }
        }

        private void mnuFileExport_Click(object sender, EventArgs e)
        {

        }

        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            if (bRunning)
            {
                SerialMon.Stop(Port);
                base.OnClosing(e);
            }

            // Save settings
            SaveSettings();
        }

        private void LoadSettings()
        {
            object oPort = Registry.GetValue(ApplicationKey, "Port", null);
            if (oPort == null)
                oPort = "COM1";
            sSerialMonPort = oPort.ToString();
        }

        private void SaveSettings()
        {
            Registry.SetValue(ApplicationKey, "Port", sSerialMonPort );
        }

        private void radioButtonHex_CheckedChanged(object sender, EventArgs e)
        {
            if (radioButtonHex.Checked)
            {
                if (chkRealTime.Checked)
                {
                    chkShowTime.Enabled = true;
                    chkShowAddress.Enabled = false;
                    chkRealTime.Enabled = true;
                }
                else
                {
                    chkShowTime.Enabled = false;
                    chkShowAddress.Enabled = true;
                    chkRealTime.Enabled = true;
                }
                displayMode = new DisplayModeHex(DataBuffer, displayProps);
            }
            else if (radioButtonText.Checked)
            {
                if (chkRealTime.Checked)
                {
                    chkShowTime.Enabled = true;
                    chkShowAddress.Enabled = false;
                    chkRealTime.Enabled = true;
                }
                else
                {
                    chkShowTime.Enabled = false;
                    chkShowAddress.Enabled = true;
                    chkRealTime.Enabled = true;
                }
                displayMode = new DisplayModeText(DataBuffer, displayProps);
            }
            else if (radioButtonFormatted.Checked)
            {
                chkShowTime.Checked = chkShowAddress.Checked = chkRealTime.Checked = false;
                chkShowTime.Enabled = chkShowAddress.Enabled = chkRealTime.Enabled = false;
                displayMode = new DisplayModeFormattedText(DataBuffer, displayProps);
            }
            else
            {
                chkShowTime.Checked = chkShowAddress.Checked = chkRealTime.Checked = false;
                chkShowTime.Enabled = chkShowAddress.Enabled = chkRealTime.Enabled = false;
                displayMode = new DisplayModeHexAndText(DataBuffer, displayProps);
            }
            displayProps.RealTime = chkRealTime.Checked;
            displayProps.ShowAddress = chkShowAddress.Checked;
            displayProps.ShowTime = chkShowTime.Checked;
            RecomputeScrollRanges();
            Invalidate(rcDisplay);
        }

        private void chkTime_CheckedChanged(object sender, EventArgs e)
        {
            if (chkRealTime.Checked)
            {
                chkShowAddress.Enabled = false;
                chkShowAddress.Checked = false;
                chkShowTime.Enabled = true;
            }
            else
            {
                chkShowAddress.Enabled = true;
                chkShowTime.Enabled = false;
                chkShowTime.Checked = false;
            }
            displayProps.RealTime = chkRealTime.Checked;
            displayProps.ShowAddress = chkShowAddress.Checked;
            displayProps.ShowTime = chkShowTime.Checked;

            RecomputeScrollRanges();
            Invalidate(rcDisplay);
        }

        private void chkShowAddress_CheckedChanged(object sender, EventArgs e)
        {
            if ( chkShowAddress.Checked )   // Can't show time and address
            {
                chkShowTime.Checked = false;
            }
            displayProps.RealTime = chkRealTime.Checked;
            displayProps.ShowAddress = chkShowAddress.Checked;
            displayProps.ShowTime = chkShowTime.Checked;
            RecomputeScrollRanges();
            Invalidate(rcDisplay);
        }

        private void chkShowTime_CheckedChanged(object sender, EventArgs e)
        {
            if (chkShowTime.Checked)   // Can't show time and address
            {
                chkShowAddress.Checked = false;
            }
            displayProps.RealTime = chkRealTime.Checked;
            displayProps.ShowAddress = chkShowAddress.Checked;
            displayProps.ShowTime = chkShowTime.Checked;
            RecomputeScrollRanges();
            Invalidate(rcDisplay);
        }

        private void ShowLine_CheckedChanged(object sender, EventArgs e)
        {
            displayProps.ShowLine1 = chkLine1.Checked;
            displayProps.ShowLine2 = chkLine2.Checked;
            RecomputeScrollRanges();
            Invalidate(rcDisplay);
        }

        private void mnuFileProperties_Click(object sender, EventArgs e)
        {
            AppProperties dlg = new AppProperties();
            dlg.Port = sSerialMonPort;

            if (dlg.ShowDialog(this) == DialogResult.OK)
            {
                sSerialMonPort = dlg.Port;
                Port.PortName = sSerialMonPort;
            }
        }
    }
    public class CLineProperties
    {
        private Font textFont;
        private Brush textBrush;
        private Brush backgroundBrush;

        public CLineProperties(Font fntText, Color clrText, Color clrTextBackground)
        {
            textFont = fntText;
            textBrush = new SolidBrush(clrText);
            backgroundBrush = new SolidBrush(clrTextBackground);
        }

        public Font TextFont
        {
            get { return textFont; }
        }

        public Brush TextBrush
        {
            get { return textBrush; }
        }

        public Brush BackgroundBrush
        {
            get { return backgroundBrush; }
        }
    };

}



/*
 
 
  
  
  
 * Interface

    X X X X X X X X  XXXXXXXX
    ^ ^ ^ ^ ^ ^ ^ ^
    | | | | | | | |
    | | | | | | | +---  Line 0/Line 1
    | | | | | | +-----  Data Type 0 - data, 1 Error, 2 Timing, 3 Sync
    | | | | | +-------  
 
 * Data
    00000001 XXXXXXXX       Line 1
    00000000 XXXXXXXX       Line 2
 * Error
    00000011 XXXXXXXX       Line 1
    00000010 XXXXXXXX       Line 2
    
 * status
 * timing
 * Commands
    * Set Baud rates
    * Set bits
    * set parity
    * set stop bits
    * do timing
    * invert signal

 * Messages
    * start
    * Baud detected/set
    * data
    * timing
    * sync
    * errors

 * GUI
    * Display as
        * Hex
        * ASCII
        * Custom (plugin)
        * Real time - with time gaps, clock - like hex, but time, not address.
    * Copy/Paste
    * Config ports
        * together
        * individually
        * Autobaud
        * Set
            * Baud
            * Bits
            * Parity
            * start/stop
 
 
    
 * TODO
    * autobaud
    * split display
    * configuration
    * port selection
 */